#include <math.h>

#include <qvariant.h>
#include <qlayout.h>
#include <qpainter.h>

#include <kwin/workspace.h>
#include <netwm.h>

#include <kstandarddirs.h>
#include <kapplication.h>
#include <kstddirs.h>
#include <kconfig.h>
#include <kwin/options.h>
#include <kpixmap.h>
#include <kdebug.h>

#include "kaptur.h"
#include "kapturbuttons.h"

using namespace KWinInternal;

static uint cornerSize   = 20;

static bool doInit                = true;
static int toolCaptionBarHeight;
static int normalCaptionBarHeight;

extern "C"
{
  Client * allocate(Workspace * ws, WId w, int tool)
  {
    return new Kaptur(ws, w, tool != 0);
  } /* allocate */
  void init()
  {
    // Nothing... Init needs Client API -> done in first client.
    doInit = true;
  } /* init */
  void reset()
  {
    // Indicates that elements are to be recalculated.
    doInit = true;
    Workspace::self()->slotResetAllClientsDelayed();
  } /* reset */
  void deinit()
  {
  } /* deinit */
}

Kaptur::Kaptur(Workspace * ws, WId w, bool tool, QWidget * parent, const char * name)
        : Client(ws, w, parent, name, WResizeNoErase),
          tool(tool),
          showStayOnTop(true),
          mainLayout(0),
          captionBarLayout(0),
          leftSpacer(0),
          captionSpacer(0)
{
  setBackgroundMode(NoBackground);

  resetLayout();

  leftButtonList.setAutoDelete(true);
  rightButtonList.setAutoDelete(true);

  connect(options, SIGNAL(resetClients()), this, SLOT(slotReset()));
} // Kaptur::Kaptur

Kaptur::~Kaptur()
{
} // Kaptur::~Kaptur

void Kaptur::resizeEvent(QResizeEvent * e)
{
  Client::resizeEvent(e);
  doShape();
  repaint();
} // Kaptur::resizeEvent

void Kaptur::captionChange(const QString &)
{
  doShape();
  repaint();
} // Kaptur::captionChange

void Kaptur::paintEvent(QPaintEvent *)
{
  int r(width());
  int b(height());
  int captionBarHeight = (tool?toolCaptionBarHeight:normalCaptionBarHeight);
  int captionBarLength = r-leftSpacer->geometry().right()-18;
  QFontMetrics fm(options->font(isActive(), tool));
  QPointArray a;

  QPainter p(this);

  if (captionSpacer->geometry().width())
  {
    KPixmap *cp = new KPixmap();
    cp->resize(captionSpacer->geometry().width(), captionBarHeight-3);
    QPainter p2(cp);
    p2.setPen(options->color(Options::TitleBar, isActive()));
    p2.setBrush(options->color(Options::TitleBar, isActive()));
    p2.drawRect(0, 0, cp->width(), cp->height());
    p2.setFont(options->font(isActive(), tool));
    p2.setPen(options->color(Options::Font, isActive()));
    p2.drawText(0, 0, cp->width(), cp->height(), AlignVCenter, caption());
    p2.end();
    p.drawPixmap(captionSpacer->geometry().left()+1, captionSpacer->geometry().top()+1, *cp);
    delete cp;
  }

  p.setPen(options->color(Options::TitleBar, isActive()));
  p.setBrush(Qt::NoBrush);
  p.drawRect(4, 4, r-8, b-captionBarHeight-3);
  p.drawLine(2, cornerSize-1, 2, 2);
  p.drawLine(2, 2, cornerSize-1, 2);
  p.drawLine(3, cornerSize-1, 3, 2);
  p.drawLine(2, 3, cornerSize-1, 3);
  p.drawLine(r-3, cornerSize-1, r-3, 2);
  p.drawLine(r-3, 2, r-cornerSize, 2);
  p.drawLine(r-4, cornerSize-1, r-4, 2);
  p.drawLine(r-3, 3, r-cornerSize, 3);
  p.drawLine(r-3, b-3, r-3, b-cornerSize-1);
  p.drawLine(r-4, b-4, r-4, b-cornerSize);
  p.setBrush(options->color(Options::TitleBar, isActive()));
  a.setPoints(4, r-captionBarLength-1, b-3, r-captionBarLength-captionBarHeight+3, b-captionBarHeight+1,
                 captionSpacer->geometry().left(), b-captionBarHeight+1, captionSpacer->geometry().left(), b-3);
  p.drawPolygon(a);
  p.drawRect(captionSpacer->geometry().right(), b-captionBarHeight+1, r-5, b-3);

  p.setPen(options->color(Options::TitleBar, isActive()).light(120));
  p.drawLine(3, cornerSize-1, 3, b-captionBarHeight+1);
  p.drawLine(cornerSize, 3, r-cornerSize-1, 3);
  p.drawLine(1, cornerSize-1, 1, 1);
  p.drawLine(1, 1, cornerSize-1, 1);
  p.drawLine(r-2, 1, r-cornerSize, 1);
  p.drawLine(r-captionBarLength-1, b-2, r-captionBarLength-captionBarHeight+2, b-captionBarHeight+1);
  p.drawPoint(2, cornerSize-1);
  p.drawPoint(cornerSize-1, 2);
  p.drawPoint(r-cornerSize, 2);

  p.setPen(options->color(Options::TitleBar, isActive()).dark(130));
  p.drawLine(2, b-captionBarHeight+1, r-captionBarLength-captionBarHeight+1, b-captionBarHeight+1);
  p.drawLine(r-4, cornerSize, r-4, b-cornerSize-1);
  p.drawLine(r-2, cornerSize-1, r-2, 1);
  p.drawLine(r-captionBarLength, b-2, r-2, b-2);
  p.drawLine(r-2, b-2, r-2, b-cornerSize);
  p.drawPoint(r-3, cornerSize-1);
  p.drawPoint(r-3, b-cornerSize);
//  p.drawLine(3, b-captionBarHeight, 3, r-captionBarLength-captionBarHeight+2);

  p.setPen(Qt::black);
  p.setBrush(Qt::NoBrush);
  p.drawRect(rect());
  p.drawLine(cornerSize, 2, r-cornerSize-1, 2);
  p.drawLine(2, b-captionBarHeight+2, r-captionBarLength-captionBarHeight+2, b-captionBarHeight+2);
  p.drawLine(r-captionBarLength-1, b-1, r-captionBarLength-captionBarHeight+2, b-captionBarHeight+2);
  p.drawLine(2, cornerSize, 2, b-captionBarHeight+2);
  p.drawLine(r-3, cornerSize, r-3, b-cornerSize-1);
  p.drawPoint(1, cornerSize-1);
  p.drawPoint(cornerSize-1, 1);
  p.drawPoint(r-2, cornerSize-1);
  p.drawPoint(r-cornerSize, 1);
  p.drawPoint(r-2, b-cornerSize);
  p.flush();
  p.end();
}

void Kaptur::doShape()
{
  calculateElements();

  int captionBarHeight = (tool?toolCaptionBarHeight:normalCaptionBarHeight);

  QRegion mask(0, 0, width(), height());

  int r(width());
  int b(height());
  captionSpacer->changeSize(captionWidth+6, captionBarHeight, QSizePolicy::Fixed, (isShade()?QSizePolicy::Fixed:QSizePolicy::Maximum));
  captionBarLayout->invalidate();
  captionBarLayout->activate();
  mainLayout->invalidate();
  mainLayout->activate();
  int captionBarLength = r-leftSpacer->geometry().right()-18;

  mask -= QRegion(cornerSize, 0, r-2*cornerSize, 2);
  QPointArray a;
  a.setPoints(4, 0, b-captionBarHeight+3, r-captionBarLength-captionBarHeight+3, b-captionBarHeight+3,
                 r-captionBarLength, b, 0, b);
  mask -= QRegion(a);
  mask -= QRegion(cornerSize-1, 0, 1, 1);
  mask -= QRegion(r-cornerSize, 0, 1, 1);
  if (!isShade())
  {
    mask -= QRegion(0, cornerSize-1, 1, 1);
    mask -= QRegion(r-1, cornerSize-1, 1, 1);
    mask -= QRegion(0, cornerSize, 2, b-cornerSize-1);
    mask -= QRegion(r-2, cornerSize, 2, b-2*cornerSize);
    mask -= QRegion(r-1, b-cornerSize, 1, 1);
  }

  setMask(mask);
} // Kaptur::doShape

void Kaptur::showEvent(QShowEvent *)
{
  doShape();
  repaint();
} // Kaptur::showEvent

void Kaptur::windowWrapperShowEvent(QShowEvent *)
{
  doShape();
  repaint();
} // Kaptur::windowWrapperShowEvent

void Kaptur::mouseDoubleClickEvent(QMouseEvent * e)
{
  if (captionSpacer->geometry().contains(e->pos()))
    workspace()->performWindowOperation(this, options->operationTitlebarDblClick());
} // Kaptur::mouseDoubleClickEvent

void Kaptur::stickyChange(bool b)
{
  emit(stkyChange(b));
} // Kaptur::stickyChange

void Kaptur::maximizeChange(bool b)
{
  emit(maxChange(b));
} // Kaptur::maximizeChange

void Kaptur::activeChange(bool)
{
  doShape();
  repaint();
} // Kaptur::activeChange

Client::MousePosition Kaptur::mousePosition(const QPoint & p) const
{
  int x = p.x();
  int y = p.y();
  int captionBarHeight = (tool?toolCaptionBarHeight:normalCaptionBarHeight);

  if (y < cornerSize)
  {
    if (x < cornerSize)
      return Client::TopLeft;
    else if (x > width()-cornerSize)
      return Client::TopRight;
    else
      return Client::Center;
  }

  if ((y > height()-4) && (x > width()-cornerSize))
    return Client::BottomRight;
  if ((x > width()-4) && (y > height()-cornerSize))
    return Client::BottomRight;

  return Client::Center;
} // Kaptur::mousePosition

void Kaptur::slotReset()
{
  resetLayout();
  repaint();
} // Kaptur::slotReset

void Kaptur::slotMaximize(int button)
{
  switch (button)
  {
    case MidButton:
      maximize(MaximizeVertical);
      break;

    case RightButton:
      maximize(MaximizeHorizontal);
      break;

    case LeftButton:
    default:
      maximize(MaximizeFull);
  }
} // Kaptur::slotMaximize

void Kaptur::toggleStayOnTop()
{
  setStaysOnTop(!staysOnTop());
} // Kaptur::toggleStayOnTop

KapturButton * Kaptur::createButton(const QString & s, QWidget * parent)
{
  KapturButton * b = 0;

  if (("Help" == s) && providesContextHelp())
  {
    b = new KapturButtonHelp(parent, "Help");
    connect(b, SIGNAL(help()), this, SLOT(contextHelp()));
  }

  else if ("Sticky" == s)
  {
    b = new KapturButtonSticky(isSticky(), parent, "Sticky");
    connect(b, SIGNAL(toggleSticky()), this, SLOT(toggleSticky()));
    connect(this, SIGNAL(stkyChange(bool)), b, SLOT(slotStickyChange(bool)));
  }

  else if ("StayOnTop" == s)
  {
    b = new KapturButtonStaysOnTop(parent, "Stay on top");
    connect(b, SIGNAL(toggleStaysOnTop()), this, SLOT(toggleStayOnTop()));
  }

  else if ("Iconify" == s && isMinimizable())
  {
    b = new KapturButtonIconify(parent, "Minimize");
    connect(b, SIGNAL(iconify()), this, SLOT(iconify()));
  }

  else if ("Maximize" == s && isMaximizable())
  {
    b = new KapturButtonMaximize(isMaximized(), parent, "Maximize");
    connect(b, SIGNAL(maximize(int)), this, SLOT(slotMaximize(int)));
    connect(this, SIGNAL(maxChange(bool)), b, SLOT(slotMaximizeChange(bool)));
  }

  else if ("Close" == s)
  {
    b = new KapturButtonClose(parent, "Close");
    connect(b, SIGNAL(closeWindow()), this, SLOT(closeWindow()));
  }

  else if ("Lower" == s)
  {
    b = new KapturButtonLower(parent, "Lower");
    connect(b, SIGNAL(lowerWindow()), this, SLOT(slotLowerWindow()));
  }

  return b;
} // Kaptur::createButton

void Kaptur::createButtons()
{
  leftButtonList.clear();
  rightButtonList.clear();

  QString buttons = options->titleButtonsLeft() + "|T" + options->titleButtonsRight();
  QPtrList<KapturButton> *buttonList = &leftButtonList;
  for (unsigned int i = 0; i < buttons.length(); ++i)
  {
    KapturButton * tb = 0;
    switch (buttons[i].latin1())
    {
      case 'S': // Sticky
        tb = createButton("Sticky", this);
        break;

      case 'H': // Help
        tb = createButton("Help", this);
        break;

      case 'I': // Minimize
        tb = createButton("Iconify", this);
        break;

      case 'A': // Maximize
        tb = createButton("Maximize", this);
        break;

      case 'X': // Close
        tb = createButton("Close", this);
        break;

      case 'T': // Stay On Top
        tb = createButton("StayOnTop", this);
        break;

      case '|':
        buttonList = &rightButtonList;
        break;
    }
    if (0 != tb)
      buttonList->append(tb);
  }
} // Kaptur::createButtons

void Kaptur::calculateElements()
{
  doInit = false;

  QFontMetrics fm(options->font(isActive(), false));
  normalCaptionBarHeight = 6+fm.height();
  QFontMetrics fm1(options->font(isActive(), true));
  toolCaptionBarHeight = 6+fm1.height();
  QFontMetrics fm2(options->font(isActive(), tool));
  captionWidth = fm2.width(caption());
} // Kaptur::calculateElements

void Kaptur::resetLayout()
{
  calculateElements();

  int buttonWidth = 12;
  int captionBarHeight = (tool?toolCaptionBarHeight:normalCaptionBarHeight);

  captionSpacer = new QSpacerItem(captionWidth, captionBarHeight, QSizePolicy::Maximum, QSizePolicy::Fixed);
  leftSpacer = new QSpacerItem(0, captionBarHeight, QSizePolicy::Expanding, QSizePolicy::Fixed);

  delete mainLayout;
  mainLayout = new QVBoxLayout(this, 0, 0);
  mainLayout->addSpacing(5);

  QHBoxLayout * centerLayout = new QHBoxLayout(mainLayout, 0, 0);
  centerLayout->addSpacing(5);
  centerLayout->addWidget(windowWrapper());
  centerLayout->addSpacing(5);

  captionBarLayout = new QHBoxLayout(mainLayout, 0, 0);
  captionBarLayout->addItem(leftSpacer);
  captionBarLayout->addSpacing(captionBarHeight-3);
  createButtons();
  captionBarLayout->setSpacing(3);
  for (QPtrListIterator<KapturButton> it(leftButtonList); it.current(); ++it)
  {
    captionBarLayout->addWidget(it.current(), 0, Qt::AlignVCenter);
    captionBarLayout->setStretchFactor(it.current(), 0);
    it.current()->setFixedSize(buttonWidth, buttonWidth);
  }
  captionBarLayout->addItem(captionSpacer);
  for (QPtrListIterator<KapturButton> it(rightButtonList); it.current(); ++it)
  {
    captionBarLayout->addWidget(it.current(), 0, Qt::AlignVCenter);
    it.current()->setFixedSize(buttonWidth, buttonWidth);
  }
  captionBarLayout->addSpacing(4);
} // Kaptur::resetLayout

void Kaptur::animateIconifyOrDeiconify(bool /* iconify */)
{
} // Kaptur::animateIconifyOrDeiconify
